"""The object insantiation and saving package.

Handles creating object instances from a parsed JSON file. It also saves the
runtime instances to a JSON file that is similar in structure to the intiially
parsed one using a custom encoding class.
"""
